/*
	This file is part of Warzone 2100.
	Copyright (C) 1999-2004  Eidos Interactive
	Copyright (C) 2005-2020  Warzone 2100 Project

	Warzone 2100 is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Warzone 2100 is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Warzone 2100; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
/** \file
 *  Some type definitions needed outside the sound library
 */

#ifndef __INCLUDED_LIB_SOUND_SOUNDDEFS_H__
#define __INCLUDED_LIB_SOUND_SOUNDDEFS_H__

enum class HRTFMode : int
{
	Unsupported = -1,
	Disabled = 0,
	Enabled = 1,
	Auto = 2 // Can be used for setting - will never be returned by sound_GetHRTFMode
};

#define MIN_VALID_HRTFMode HRTFMode::Disabled
#define MAX_VALID_HRTFMode HRTFMode::Auto

#endif	// __INCLUDED_LIB_SOUND_SOUNDDEFS_H__
