/*
	This file is part of Warzone 2100.
	Copyright (C) 2024  Warzone 2100 Project

	Warzone 2100 is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Warzone 2100 is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Warzone 2100; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#pragma once

#include "lib/widget/widget.h"
#include "lib/widget/form.h"

#include <functional>

#include "lib/netplay/netjoin.h"	// for JoinConnectionDescription
#include "lib/framework/crc.h"

struct ExpectedHostProperties
{
	optional<EcKey::Key> hostPublicKey;
	optional<std::string> gamePassword;
};

namespace netlobby
{
	struct ConnectionType;
}

bool startLobbyJoiningAttempt(char* playerName, const std::string& lobbyAddress, const std::string& gameId, bool asSpectator = false, const ExpectedHostProperties& expectedHostProps = ExpectedHostProperties(), const std::vector<netlobby::ConnectionType>* pKnownAvailableConnectionTypes = nullptr);

bool startJoiningAttempt(char* playerName, std::vector<JoinConnectionDescription> connection_list, bool asSpectator = false, const ExpectedHostProperties& expectedHostProps = ExpectedHostProperties());
bool startJoinRedirectAttempt(char* playerName, std::vector<JoinConnectionDescription> connection_list, bool asSpectator = false, const ExpectedHostProperties& expectedHostProps = ExpectedHostProperties());
void resetJoinRedirectTracking();
void shutdownJoiningAttempt();

std::shared_ptr<WIDGET> createJoiningIndeterminateProgressWidget(iV_fonts fontID);
