// SPDX-License-Identifier: GPL-2.0-or-later

/*
	This file is part of Warzone 2100.
	Copyright (C) 2025  Warzone 2100 Project

	Warzone 2100 is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Warzone 2100 is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Warzone 2100; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "tcp_address_resolver.h"

#include "lib/netplay/tcp/netsocket.h"
#include "lib/netplay/tcp/tcp_connection_address.h"

#include <memory>

namespace tcp
{

TCPAddressResolver::TCPAddressResolver()
{
	SOCKETinit();
}

TCPAddressResolver::~TCPAddressResolver()
{
	SOCKETshutdown();
}

net::result<std::unique_ptr<IConnectionAddress>> TCPAddressResolver::resolveHost(const char* host, uint16_t port) const
{
	auto resolved = tcp::resolveHost(host, port);
	if (!resolved.has_value())
	{
		return tl::make_unexpected(resolved.error());
	}
	return std::make_unique<TCPConnectionAddress>(resolved.value());
}

} // namespace tcp
